//
//
//

#include "cordnum.h"

#include <lpmd/matrix.h>
#include <lpmd/util.h>
#include <lpmd/simulation.h>

#include <sstream>

using namespace lpmd;

CordNum::CordNum(std::string args): Plugin("cordnum", "2.0")
{
 trcut = 0.0;
 ParamList & params = (*this);
 //
 DefineKeyword("atoms");
 DefineKeyword("start", "0");
 DefineKeyword("end", "-1");
 DefineKeyword("each", "1");
 DefineKeyword("rcut");
 DefineKeyword("maxn");
 DefineKeyword("output");
 ProcessArguments(args);
 nb = int(params["maxn"]);
 start = int(params["start"]);
 end = int(params["end"]);
 each = int(params["each"]);
 OutputFile() = params["output"];
}

CordNum::~CordNum()
{
}

void CordNum::SetParameter(std::string name)
{
 //#warning "SetParameter es horrible! mejorar los parametros"
 if (name == "atoms")
 {
  AssignParameter("atoms", GetNextWord());
  na = int((*this)["atoms"]);
  for(int i=0;i<na;i++) { satoms.push_back(GetNextWord()); }
 }
 else if (name == "rcut") 
 {
  std::string atom1 = GetNextWord();
  std::string atom2 = GetNextWord();
  std::string tmp = GetNextWord();
  double cf = atof(tmp.c_str());
  rcut[atom1+"-"+atom2] = cf;
  rcut[atom2+"-"+atom1] = cf;
  trcut += cf;
 }
 else Module::SetParameter(name);
}

void CordNum::Show(std::ostream & os) const
{
 Module::Show(os);
 if(na > 0)
 {
  std::cout << "   Atoms N     = " << na << '\n';
  std::cout << "   Atoms       = ";
  for(unsigned int i=0;i<satoms.size();i++) std::cout << satoms[i] << "\t";
  std::cout << std::endl;
  std::cout << "   Max neigh   = " << nb << '\n';
  std::cout << "   Cutoffs     = " << '\n';
  for(unsigned int i=0;i<satoms.size();i++)
  {
   for(unsigned int j=i;j<satoms.size();j++)
   {
    std::string spec1=satoms[i];
    std::string spec2=satoms[j];
    std::string tmp = spec1+"-"+spec2;
    // Truco para acceder a rcut[tmp] desde un metodo const 
    const std::map<std::string, double>::const_iterator & p = rcut.find(tmp);
    double cut = (*p).second;
    // fin del truco
    std::cout <<"\t"<< spec1 << "-" << spec2 << " = " << cut << std::endl;
   }
  }
 }
}

void CordNum::ShowHelp() const
{
 std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
 std::cout << " Module Name        = cordnum                                                  \n";
 std::cout << " Problems Report to = admin@lpmd.cl                                            \n";
 std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
 std::cout << " General Info      >>                                                          \n";
 std::cout << "      This plugin is used to calculate the coordination number (CN) of an      \n";
 std::cout << "      atomic configuration. The CN is calculated between the neighbors of the  \n";
 std::cout << "      subcells generated by the linked cell method. This plugin generates a    \n";
 std::cout << "      histogram, and is not the usual CN (see also 'cordnumfunc' plugin).      \n";
 std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
 std::cout << " General Options   >>                                                          \n";
 std::cout << "      maxn          : Sets the maximum number of neighbors to write the        \n";
 std::cout << "                      histogram, usually, more than 10.                        \n";
 std::cout << "      atoms         : Sets the number of atomic species and their symbols for  \n";
 std::cout << "                      the calculation of the CN.                               \n";
 std::cout << "      rcut          : Sets the two species and cutoff radius.                  \n";
 std::cout << "      output        : Output file.                                             \n";
 std::cout << "      average       : Sets if the the property must be averaged over all       \n";
 std::cout << "                      configurations (true / false)                            \n";
 std::cout << "      cutoff        : Cutoff radius for the neighbors list. If rcut is given,  \n";
 std::cout << "                      the default value is 2*rcut                              \n";
 std::cout << "      start         : Determines in which step the plugin begins to be applied.\n";
 std::cout << "      end           : Determines in which step the plugin ceases to be applied.\n";
 std::cout << "      each          : Determines how often (each how many time-steps) the      \n";
 std::cout << "                      plugin must be applied.                                  \n";
 std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
 std::cout << " Example           >>                                                          \n";
 std::cout << " #Loading the plugin :                                                         \n";
 std::cout << " use cordnum                                                                   \n";
 std::cout << "     maxn 15                                                                   \n";
 std::cout << "     atoms 1 Ar                                                                \n";
 std::cout << "     rcut Ar Ar 3.95                                                           \n";
 std::cout << "     output cordnum.dat                                                        \n";
 std::cout << "     average false                                                             \n";
 std::cout << "     cutoff 9                                                                  \n";
 std::cout << " enduse                                                                        \n";
 std::cout << " #Applying the plugin :                                                        \n";
 std::cout << " property cordnum start=0 each=1 end=100                                     \n\n";
 std::cout << "      The plugin is used to calculate the CN between argon (Ar) atoms every    \n";
 std::cout << "      step of the first 100 steps. The data is written in the file cordnum.dat \n";
 std::cout << "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
}

void CordNum::Evaluate(Configuration & con, Potential & pot)
{
 assert(&pot != 0);//icc 869
 lpmd::BasicParticleSet & atoms = con.Atoms();
 lpmd::Array <int> esp = atoms.Elements();
 if (nb <= 0 || na <=0) throw PluginError("cordnum", "Error in coordination number calculation.");
 int nsp = na;
 unsigned long int N = atoms.Size();
 int **histo;
 double **cnfun;
 histo = new int*[nsp*nsp];
 cnfun = new double*[nsp*nsp];
 for(int i=0;i<nsp*nsp;i++) { histo[i]=new int[N]; }
 for(int i=0;i<nsp*nsp;i++) { cnfun[i]=new double[nb];}
 for (int i=0;i<nsp*nsp;i++)
 { 
  for (unsigned long int j=0;j<N;j++) histo[i][j]=0;
  for (int j=0;j<nb;j++) cnfun[i][j]=0.0e0;
 }

 //const std::list<std::string> lst = simcell.RepeatedSpeciesPairs(); 
 lpmd::Array<std::string> pairs;
 for(int i=0; i<esp.Size() ; ++i)
 {
  for(int j=0; j<esp.Size() ; ++j)
  {
   std::ostringstream ostr;
   ostr << ElemSym[esp[i]]<< "-" << ElemSym[esp[j]];
   pairs.Append(ostr.str());
  }
 }
 int s=0;

 for(int i=0;i<pairs.Size();++i)
 {
  //Hace funcional cada una de las especies de los pares.
  lpmd::Array<std::string> loa = StringSplit(pairs[i],'-'); // lista de atomos
  int e1 = ElemNum(loa[0]);
  int e2 = ElemNum(loa[1]);
  double rc12 = rcut[loa[0]+"-"+loa[1]];
  //Cuenta los atomos de la especie 1.
  int ne1=0;
  for (unsigned long int k=0;k<N;++k) {if(atoms[k].Z()==e1) ne1++;}
  //Comienzan las iteraciones.
  if (fabs(trcut)<1e-1) trcut = rc12*2;
  for (unsigned long int j=0;j<N;++j)
  {
   if(atoms[j].Z()==e1)
   {
    lpmd::NeighborList & nlist = con.Neighbors(j,true,trcut);
    for(long int k=0;k<nlist.Size();++k)
    {
     const lpmd::AtomPair & nn = nlist[k];
     if(nn.j->Z()==e2)
     {
      if(nn.r2<=rc12*rc12)
      {
       histo[s][j]++;
      }
     }
    }
   }
  }
  for (unsigned long int j=0;j<N;j++)
  {
   if(atoms[j].Z()==e1 && histo[s][j]<nb)
   {
    cnfun[s][histo[s][j]]++;
   }
   histo[s][j]=0;
  }
  double distp=0.0e0;
  for(int j=0;j<nb;j++)
  {
   distp+=j*cnfun[s][j];
   cnfun[s][j]=(cnfun[s][j]/ne1);
  }
  s++;
 }
 //
 // Output of cordnum - histogram format
 //
 lpmd::Matrix & m = CurrentValue();
 m = lpmd::Matrix(1 + nsp*nsp, nb);
 // Asigna los labels al objeto Matrix para cada columna
 m.SetLabel(0, "numb of neigh");
 int j=1;
 for (int i=0;i<pairs.Size();++i)
 {
  m.SetLabel(j, pairs[i]+" cn");
  j++;
 }
 //
 for(int i=0;i<nb;i++)
 {
  m.Set(0, i, i);
  for(j=0;j<(int)(nsp*nsp);j++)
  {
   m.Set(j+1, i, cnfun[j][i]);
  }
 }
 //Borra arreglos dinamicos.
 for(int i=0;i<nsp*nsp;i++)
 {
  delete[] cnfun[i];
  delete[] histo[i];
 }
 delete [] cnfun;
 delete [] histo;
}

// Esto se incluye para que el modulo pueda ser cargado dinamicamente
Plugin * create(std::string args) { return new CordNum(args); }
void destroy(Plugin * m) { delete m; }

